--------------------------------- ansic.m4 -----------------------------------
   This is the m4 include file that defines enough stuff so that
C is turned into an instance of our generic assembler.


Provide({ansic})



------------------------------------------------------------------------------
----------------------------- Small C library --------------------------------
------------------------------------------------------------------------------


---------------------- Provide defaults for some constructs ------------------

defdef({__inline__})


------------------------------ Basic Types -----------------------------------

define({__C_BASIC_TYPES__},{
typedef void*	ptr_t;
})


------------------------ Memory access routines ------------------------------


define({DefineXXAccess},{
inline void put_$1(ptr_t x,$1 y)
BEGIN
  *(($1*)x)=y ;
END
inline Byte get_$1(ptr_t x)
BEGIN
  return *(($1*)x) ;
END
})






----------------------------- non-local exits --------------------------------

__NLEXIT_HEADER__	macro that prepares a loop for non local exit
__NLEXIT_FOOTER__	macro that define variables so as to allow further exit
__NLEXIT_TEST__	macro to test for non-local exit
__NLEXIT_GO__	macro to do a non-local exit
	all these take one argument as unique name/prefix of the structure
	being defined for non-local exit.
__NLEXIT_localonly__ is defined iff the non-locality is emulated and needs
		functions to return to a local scope whenever they issued
		such an exit, and to skip things that an actual non-local
		exit would skip.


ifdef({HAS_SETJUMP},{

define({__NLEXIT_HEADER__},{
	DeclareVariable($1__exit_buf,jmp_buf)
	if(!setjmp($1__exit_buf))
	BEGIN
})
define({__NLEXIT_FOOTER__},{
	END
})
define({__NLEXIT_TEST__},{
	BEGIN ; END
})
define({__NLEXIT_GO__},{longjmp($1__exit_buf,-1);})

},{

define({__NLEXIT_HEADER__},{
	DeclareVariable($1__exit_p,boolean)
	BEGIN
	  $1__exit_p = FALSE;
})
define({__NLEXIT_FOOTER__},{
	END
	$1__exit_label:
})
define({__NLEXIT_TEST__},{
	BEGIN if ($1__exit_p) goto $1__exit_label ; END
})
define({__NLEXIT_GO__},{$1__exit_p = TRUE;})

define({__NLEXIT_localonly__})

})



------------------------------------------------------------------------------
------------------------- Define Pseudo-ASM machine --------------------------
------------------------------------------------------------------------------


------------------------------- General syntax -------------------------------

define({Terminator},{;})
define({Separator},{})



--------------------------------- Structures ---------------------------------





--------------------------- Pseudo-ASM registers -----------------------------

defdef({PC},	{PC})dnl	-- program counter
defdef({DSP},	{DSP})dnl	-- data stack pointer
defdef({DSINC},	{(-CELL})dnl	-- increment for data stack, grow down
defdef({TOS},	{TOS})dnl	-- buffered top of stack
defdef({TOSL},	{LowWord(TOS)})dnl	-- its lowest word
defdef({TOSLL},	{LowByte(TOS)})dnl	-- its lowest byte
defdef({TOSLH},	{SecondByte(TOS)})dnl	-- its second byte
defdef({RSP},	{RSP})dnl	-- return stack pointer
defdef({RSINC},	{(-CELL)})dnl	-- increment for return stack -- grow down
defdef({FRAME},	{})dnl	-- a frame pointer for heap manipulation
defdef({R0},	{})dnl	-- a scratch register
defdef({R0L},	{})dnl	-- its lowest word
defdef({R0LL},	{})dnl	-- its lowest byte
defdef({R0LH},	{})dnl	-- its second byte
defdef({R1},	{})dnl	-- second scratch register
defdef({COUNT},	{COUNT})dnl	-- count register
defdef({COUNTLO},{LowByte(COUNT)})dnl	-- count register -- low byte


dnl --- Define Pseudo-ASM type-casting ---

define({bytecast},{((Byte)$1)})
define({wordcast},{((Word)$1)})
define({cellcast},{((Cell)$1)})


dnl --- Define Pseudo-ASM pointers ---

define({genpoint},
{ifelse({$3},{},
  {(*(($1*) ($2)))},
  {(*( (($1*) ($2)) + ($3) ))})})

define({point},{genpoint({Cell},$@)})
define({bytepoint},{genpoint({Byte},$@)})
define({wordpoint},{genpoint({Word},$@)})


dnl --- Define Pseudo-ASM Instructions ---

define({Mov},{$1=$2})dnl
define({Add},{$1+=$2})dnl
define({Sub},{$1-=$2})dnl
.......


dnl --- Define Pseudo-ASM structure ---

define({Loop},{for(;COUNT!=0;COUNT--)  BEGIN  $1  END})
define({exit_Loop},{break;})




dnl
dnl --------------------------- Structures --------------------------------
dnl
dnl

struct API_Signature {
   int32 Magic_Number ;
   byte  Object_Name[20] ;
   byte  Version[8] ;
};

struc API_Header {
   int32 TotalLength ;		/* total length of the file */
   int32 EntryPoint ;		/* address of object to return */
   int32 Offset	;		/* offset where to load it */
   int32 HeaderSize ;		/* size of header fields in file */
   int32 ExpandedSize ;		/* size when expanded to memory */
   int32 ___ ;			/* fill in */
   int32 CheckSum ;		/* checksum */
   int32 LongCheckSum ;		/* long checksum */
})


dnl
dnl ---------------------- C as an assembler -------------------------------
dnl
dnl


define({CELL},	CELLSIZE)		; a cell is 4 bytes long


dnl
dnl ---- abstraction of the Pseudo-ASM instruction set ----
dnl
define({Add},		{($1) += ($2) ;})dnl
define({Sub},		{($1) -= ($2) ;})dnl
define({__Mov__},	{($1) = ($2) ;})dnl
define({__Movb__},	{($1) = ($2) ;})dnl
define({__Movw__},	{($1) = ($2) ;})dnl
define({Neg},		{($1) = - ($1) ;})dnl
define({Xchg},		{R0 = $1 ; $1 = $2 ; $2 = R0 ;})dnl
define({},		{})dnl
define({},		{})dnl
dnl
dnl
define({point},	{ifelse({$2},{},{*((cell*)$1)},{((cell*)$1)[$2]})})dnl
dnl
dnl
define({PC},	{esi})dnl	program counter
define({DSP},	{edi})dnl	data stack pointer
define({DSINC},	{CELL})dnl	increment for data stack -- grow up
define({TOS},	{ebx})dnl	buffered top of stack
define({RSP},	{esp})dnl	return stack pointer
define({RSINC},	{-CELL)})dnl	increment for return stack -- grow down

define({FRAME},		{ebp})	; a frame pointer for heap manipulation
define({R0},		{eax})	; a scratch register
define({R0L},		{al})	; its lowest byte
define({R0H},		{ah})	; its second byte
define({R1},		{edx})	; second scratch register
define({COUNT},		{ecx})	; count register
define({COUNTLO},	{cl})	; count register -- low byte

define({INSTRTBL},	InstrT)		; table for bytecode threading

define({PopData},	{Mov({$1},TOS)
			 Add(DSP,-DSINC)})
define({PushData},	{Add(DSP,DSINC)
			 Mov(point(DSP),TOS)
			 Mov(TOS,{$1})})
define({PopReturn},	{ifelse({$1},,{Add(RSP,-RSINC)},{pop $1})})
define({PushReturn},	{push $1})
define({},		{})
define({},		{})
define({},		{})
define({},		{})

define({Enter},	{Add(PC,CELL)})
define({Next},	{jmp [PC]})


;;;;;;;;;;;;;;;;;;;;;;;;;; Defining FORTH words ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; defining a counted string
define({DefCString},
{__stringseg__({
locals({strlen},{strbeg},{strend},{
strbeg:
	.byte 	strlen
	defb($@)
strend:
strlen	=	strend-strbeg-1
})})})

;; defining a FORTH word


define({DefWord},{
ifelse({$2},,{$1},{$2}):
})






/*** Some instructions ***/

define({DefCompareWord0},
{DefWord({$1},{$2},{$3})
   Enter
   TOS=(TOS $4 0)?FALSE:TRUE ;
   Next
})dnl
define({DefCompareWord},
{DefWord({$1},{$2},{$3})
   Enter
   TOS=(NOS $4 TOS)?FALSE:TRUE ;
   Nip
   Next
})

