dnl ---------------------=|=--- paging.m4 ---=|=-----------------------------
dnl --=-- paginating virtual memory manager for the TUNES i386 API --=--
dnl
dnl --- See ~moose/info/BIOS-API.nfo for more info about this. ---
dnl
dnl Beware: must be included at the end and/or be the last module
dnl to have data in udataseg !!!
dnl
Provide({paging},{
dnl
dnl

pmpart

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

InitialTopFreePages	= (pmaddr(pmpart_actbeg)) & PAGEMASK

Top_Pgtbl		= InitialTopFreePages-pmofs
; Ahem this is a dirty optimization. See Init_Paging to remove it.

K_CR3			= K_TSS+TSS32_CR3
; CR3 for Kernel


zzz_Init_Hole_Size0	= pmpart_beg-pmpart_actbeg
zzz_Init_Hole_Size	= pmpart_beg-Top_Pgtbl
zzz_pmpart_actsize	= pmpart_end-Top_Pgtbl+8192
; just to see how long it is in biosapi.l symbol table


__dataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Variables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Address of the GDT descriptor
GDT_Addr_Addr:	.long	rmaddr(PM_GDT_address)	; (to adjust)


; .align 4
; Id_Pgtbl_Sz:	.long	0
Id_Pgtbl_Num:	.long	0	; number of pages to id-map physical memory

K_PageMap:	.long	-0x400000 ; -4MB-space to map physical memory

; This is needed because the GDT points to virtual memory, not real
; (or doesn't it), so if we are to remap low linear addresses, we need
; to reload the gdt first !
PG_GDT_address:
	PG_GDT_lim:	.word	GDT_SIZE-1
	PG_GDT_base:	.long	pmaddr(__GDT__)
PG_IDT_address:
	PG_IDT_lim:	.word	IDT_SIZE-1
	PG_IDT_base:	.long	pmaddr(__IDT__)


; see Init_Paging
; it contains the linear address of the page directory structure:
; 4MB of linear memory that maps automagically to the mapping tables !
; (trick taken from VSTa)




;;; (!) Variables a Relocator GC shall be aware of:


})


__udataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;; Uninitialized Data ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; (!) Variables a Relocator GC shall be aware of:




;;; Normal Variables

;Paging_Flags		= allot(u,1)
;Paging_on		= 0x01	; is paging on ?
;Paging_id		= 0x02	; is paging identity mapping physical memory ?

Top_Pgtbl_Num		= alallot(u,4)
; number of pages to map
; the physical memory at bottom and top of linear memory respectively
})


__pmudataend__({

;;; (!) This must be at the very end: page-alignment !!!
;;; perhaps we should create a new segment for that, a "pu" heap ?
alignheap(u,16)

;; Permanent:
; * K_PageDir is the "Kernel" page directory for the "Kernel" task,
;  that identity-maps physical memory.
;  Such task is necessary for isolation boxes, even though coarse-grained
;  tasks are not the "usual" way of doing multithreading.
; * Null_Page is a page full of zeroes for read-only/copy-on-write pages
;  it is zeroed during bss initialization.

K_PageDir		= allot(u,PAGESIZE)
Null_Page		= allot(u,PAGESIZE)


;; Testing:
; * T_PageDir is a "Test" page directory
; * T_Pgtbl is a "Test" page table
T_PageDir		= allot(u,PAGESIZE)
T_Pgtbl			= allot(u,PAGESIZE)

;; (!) must be the very last thing in the udataseg !!!
Id_Pgtbl		= allot(u,0)

})



__codeseg__({

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;; Switching between PL0 and PLx ;;;;;;;;;;;;;;;;;;;;;;

PL0toPL1:
  pushm(#KL1DS,#KL1CS)

PL0toPLx:
  popm(es,ds)
  pushm(esp,ds)		; PLx stack
  add [esp],8
  pushm([esp+12],es)	; PLx target
  ppmov(es,ds)		; PLx data segments
  zzzz({ppmov(fs,#TEMPSEG+3)})
  retf			; jump to PLx

PL0toPL3:
  pushm(#KL3DS,#KL3CS)
  jmp PL0toPLx


PL1toPL0:
;;; Getting back from PL1 to PL0:
  pushing({fd},{
  cli
  calli #0,KL1CG
KL1_CallGate:
  mov esp,[esp+8]
  })
  ppmov(ds,#KL0DS)
  ppmov(es,ds)
  ppmov(ss,ds)
  zzzz({ppmov(fs,ds)})
  ret


KL3_CallGate:
;;; System call manager
;;Input:
;;Output:
	retf 32 ; 8 arguments


KL0_CallGate:
KL2_CallGate:
	cli
	___die___ ; these should never be used


;;;;;;;;;;;;;;;;;;;;;;;;;;;; Switching Pagination ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Enable_Paging:
;;Input:
;; [esp]: near continuation
;; cli assumed, as well as being in PL0, using KL0 segments in cs,ds,es,ss
;;Output:
;; Paging enabled
;; gdt, idt reloaded
;; cs,ds,es,ss initialized up in linear memory.
;; eax,esi wasted
;;Question:

	STOP(fs,0xB8000,0,23,'E,0x18)

   ;; User space is identity mapped to physical memory
	ttmov(cr3,[pmaddr(K_CR3)],eax)

   ;; Do enable
	mov eax,cr0
	or eax,#CR0_PG
	mov cr0,eax
	STOP(fs,0xB8000,1,23,'P,0x18)

   ;; Reload gdt, idt
	lgdt [pmaddr(PG_GDT_address)]
	lidt [pmaddr(PG_IDT_address)]

   ;; Take care of segments
	mov esi,#pmaddr(__GDT__)+ALLMEMCS
	;jmp Commonable_Paging


Commonable_Paging:
;;Input:
;; esi: base of descriptors to copy to KL0 descriptors
;;Output:
;; eax is wasted
	STOP(fs,0xB8000,2,23,'C,0x18)
   ;; Modify KL0xS entries in GDT
	; (!) hackish optimizations:
	; first dword (low of limit, low of base) is same for KL0CS and KL0DS
	; second dword is same except AH=0x9A for KL0CS, AH=0x92 for KL0DS
	mov eax,[esi]
	mov [pmaddr(__GDT__)+KL0CS],eax
	mov [pmaddr(__GDT__)+KL0DS],eax
	mov eax,[esi+4]
	mov [pmaddr(__GDT__)+KL0CS+4],eax
	mov ah,#0x93
	mov [pmaddr(__GDT__)+KL0DS+4],eax
   ;; Reload segment registers to use new values
	movzil(a,#KL0DS)
	mov ds,ax
	mov es,ax
	mov ss,ax
	push [esp]
	mov [esp+4],#KL0CS
	retf


Disable_Paging:
;;Input:
;; cli assumed
;;Output:
;;; Paging disabled
;; segments initialized up in physical memory.
;; eax,esi wasted
	STOP(fs,0xB8000,0,23,'D,0x018)
   ;; User space is identity mapped to physical memory
	ttmov(cr3,[pmaddr(K_CR3)],eax)
   ;; Take care of segments
	mov esi,#pmaddr(__GDT__)+PMPARTCS
	STOP(fs,0xB8000,1,23,'p,0x18)
	call Commonable_Paging
   ;; Do disable
	mov eax,cr0
	and eax,#~CR0_PG
	mov cr0,eax
   ;; Reload gdt, idt
	mov eax,[pmaddr(GDT_Addr_Addr)]
	lgdt [eax]
	lidt [pmaddr(PM_IDT_address)]
   ;; Done
	ret


})

pminit

__codeseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;; Initialization routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;


Init_Paging:
;;Input:
;;; TopFreeMem must be uninitialized
;; cld assumed
;;Output:
;;; Top_Pgtbl is statically allocated
;;; TopFreeMem is *initialized*
;;; everything is ready for paging to take place.
;;; paging is enabled
;; cld forced
;; all gregisters saved, *but eax*

	STOP(fs,0xB8000,0,24,'P,0x09)
	pushad

;;; Firstly, initialize some individual fields
	mov ebp,[pmaddr(PM_Part_Phys)]
	lea eax,[ebp+pmaddr(K_PageDir)]
	mov [pmaddr(K_CR3)],eax

;;; Initialize mapping table of top of linear memory to physical memory
;;; so we have the same address in kernel as anywhere else.
   ;; Get address of top of table
	;mov edi,[pmaddr(TopFreeMem)] ; (!) here the dirty optimization
	mov edi,#pmaddr(Top_Pgtbl)-4
	STOP(fs,0xB8000,1,24,'A,0x09)


   ;; Get address of top of memory
   ;; what's physically above this address is above linear address 0xFFFFFFF
   ;; in kernel space, and thus not visible from kernel space
   ;; (!) we assume PM_Part_Phys is page-aligned - see head.m4
	mov eax,ebp
	STOP(fs,0xB8000,2,24,'T,0x09)

   ;; Adjust address into a valid page table entry
   ;; (!) beware: PG_US means user-level can access it
   ;; That's temporary, for testing.
   ;; Well, isolation boxes should already prevent the user from
   ;; accessing it through segmentation, so it should be safe anyway
	add eax,#PG_P|PG_RW|PG_US-PAGESIZE
	STOP(fs,0xB8000,3,24,'A,0x09)

   ;; Do initialize
	std
     Init_Paging_0:
	stosd
	sub eax,#PAGESIZE
	jae Init_Paging_0
	STOP(fs,0xB8000,4,24,'I,0x09)

   ;; Update TopFreeMem, Top_Pgtbl_Num, K_PageMap
	add edi,#4
	and edi,#PAGEMASK
	mov [pmaddr(TopFreeMem)],edi
	mov eax,#pmaddr(Top_Pgtbl) ; (!) same dirty opt. as above
	sub eax,edi
	shr eax,#PAGEBITS
	mov [pmaddr(Top_Pgtbl_Num)],eax
	mov ecx,eax	; (!) see below
	shl eax,#22
	sub [pmaddr(K_PageMap)],eax
	STOP(fs,0xB8000,4,24,'T,0x09)

   ;; enable reverse mapping in "Kernel" page directory
	;mov ecx,[pmaddr(Top_Pgtbl_Num)]	; assumed from above
	mov edi,#pmaddr(K_PageDir)+PAGESIZE-4
	mov eax,#pmaddr(Top_Pgtbl)-PAGESIZE+PG_P|PG_RW|PG_US
	add eax,ebp
	 ; (!) same dirty opt. as above
	 ; present RW pages allow the automagical PageMap trick
	 ; same warning for PG_US
	;std ; assumed from above
	Loopno0({
	stosd
	sub eax,#PAGESIZE
	})
	STOP(fs,0xB8000,5,24,'E,0x09)


;;; Initialize identity mapping tables
   ;; Get number of pages to map
   ;; (!) we use XtdMem
	mov ecx,[pmaddr(XtdMem)]
	add ecx,#0x100000+PAGESIZE-1	; XtdMem is memory above 1MB
	shr ecx,#PAGEBITS		; we get the number of pages
	clr(edi) ;mov edi,#pmaddr(Id_Pgtbl) ; (!) optim: Id_Pgtbl is at zero
	STOP(fs,0xB8000,0,24,'P,0x2A)

   ;; make a valid page table entry
     ; (!) beware: PG_US means user-level can access it
     ; That's temporary, for testing.
     ; Well, isolation boxes should already prevent the user from
     ; accessing it through pagination, so it should be safe anyway
	mov eax,#PG_P|PG_RW|PG_US
	STOP(fs,0xB8000,1,24,'a,0x2A)

   ;; Do initialize
	cld
	Loopno0({
	stosd
	add eax,#PAGESIZE
	})
	STOP(fs,0xB8000,2,24,'t,0x2A)

   ;; (!) Note that this space was reserved when relocating the kernel
   ;; to top of memory

   ;; enable identity mapping in "Kernel" page directory
	mov ecx,[pmaddr(Id_Pgtbl_Num)]
	mov edi,#pmaddr(K_PageDir)
	mov eax,#pmaddr(Id_Pgtbl)+PG_P|PG_RW|PG_US
	add eax,ebp
	 ; present RW pages allow the automagical PageMap trick
	 ; same warning as above for PG_US
	;cld ; assumed from above
	Loopno0({
	stosd
	add eax,#PAGESIZE
	})
	STOP(fs,0xB8000,3,24,'a,0x2A)

   ;; (!) Initialize test page directory
zzz({
	mov esi,#pmaddr(K_PageDir)
	mov edi,#pmaddr(T_PageDir)
	mov ecx,#PAGESIZE/4
	pushing({ecx},{ rep
			movsd})
	STOP(fs,0xB8000,4,24,'t,0x2A)
	mov esi,#pmaddr(Id_Pgtbl)
	mov edi,#pmaddr(T_Pgtbl)
	rep
	movsd
	mov [pmaddr(T_Pgtbl)+4],#0x000B8000|PG_P|PG_RW|PG_US
	STOP(fs,0xB8000,5,24,'e,0x2A)
	;.....
})

;; Enable paging and return
	popad
	STOP(fs,0xB8000,7,24,' !,0x09)
	br Enable_Paging

})


})dnl --- end of paging.m4
dnl
